;-------------------------------------------------
;	snd_fmt.bb
;-------------------------------------------------
;	The included function 'GetAudioFormat' shows	
;	users how to create a process each codec and	
;	generate a list of formats available on the	
;	target computer.									
;-------------------------------------------------
;	*NOTE* If you want to give users the easiest
;	way to select a format, use snd_format_choose.
;	These methods are available incase you want to
;	choose something without user interface, etc.
;--------------------------------------------------
;	by Nullmind
;--------------------------------------------------
;	Todo:
;		- Create ways to compare information from
;		  loaded codecs with installed codecs.
;		- Adapt to later BlitzPlus versions for
;		  better TreeView support. (B+ is ghetto)

br = Chr(10)
format = GetAudioFormat("Playback Format", 0)

Type format_tag
	Field name$
	Field tag
	Field index
End Type

Type format
	Field tag
	Field format
	Field name$
	Field index
End Type

Function GetFormatTag.format_tag(tag%)
	For ft.format_tag = Each format_tag
		If(ft\tag = tag) Return(ft)
	Next
	
	Return(Null)
End Function

Function GetFormatTagByIndex.format_tag(index%)
	For ft.format_tag = Each format_tag
		If(ft\index = index) Return(ft)
	Next
	
	Return(Null)
End Function

Function GetFormatByIndex.format(index%)
	For tf.format = Each format
		If(tf\index = index) Return(tf)
	Next
	
	Return(Null)
End Function

Function FormatAlreadyExists(format%)
	For tf.format = Each format
		If(snd_format_tag(format) = snd_format_tag(tf\format))
			If(snd_format_channels(format) = snd_format_channels(tf\format))
				If(snd_format_samples(format) = snd_format_samples(tf\format))
					If(snd_format_bits(format) = snd_format_bits(tf\format))
						If(snd_format_bps(format) = snd_format_bps(tf\format)) Return(True)
					EndIf
				EndIf
			EndIf
		EndIf
	Next
	
	Return(False)
End Function

Function ProcessFormats(flags%=0)
	Local index = 0
	Local codecs
	
	codecs = snd_codec_count()
	For c=0 To codecs-1
		formats = snd_codec_formats(c, flags)
		For f=0 To formats-1
			Local format, tft.format_tag, tag, tf.format
			
			format = snd_codec_getformat(f)
			tag = snd_format_tag(format)
			
			If(GetFormatTag(tag)=Null)
				tft = New format_tag
				tft\tag = tag
				tft\name = snd_format_tagname(tag)
				tft\index = index
				index = index + 1
			EndIf
			
			If(FormatAlreadyExists(format))
				snd_format_free(format)
			Else
				tf.format = New format
				tf\tag = tag
				tf\format = format
				tf\name = snd_codec_formatname(f)
			EndIf
		Next
	Next
End Function

Function GetAudioFormat(title$="Audio Format", flags=0, parent=0)
	If(parent = 0) parent = Desktop()
	Local x = GadgetX(parent) + GadgetWidth(parent)/2 - 256/2
	Local y = GadgetY(parent) + GadgetHeight(parent)/2 - 412/2
	Local window = CreateWindow(title, x, y, 256, 412, 0, 1+2)	
	Local tree = CreateTreeView(4, 4, ClientWidth(window)-8, ClientHeight(window)-8-26-4, window)
	Local accept = CreateButton("Accept", ClientWidth(window)-80-4, ClientHeight(window)-26-4, 80, 26, window)
	Local format
	
	SetGadgetLayout(tree, 1, 1, 1, 1)
	SetGadgetLayout(accept, 0, 1, 0, 1)
	
	ProcessFormats(flags)
	
	For tft.format_tag = Each format_tag
		tft\index = AddTreeViewNode(tft\name, TreeViewRoot(tree))
		
		For tf.format = Each format
			If(tf\tag = tft\tag)
				tf\index = AddTreeViewNode(tf\name, tft\index)
				
				If(SelectedTreeViewNode(tree) = 0)
					SelectTreeViewNode(tf\index)
				EndIf
			EndIf
		Next
	Next
	
	While(WaitEvent())
		Select EventID()
			Case $401
				Select EventSource()
					Case tree
						If(GetFormatTagByIndex(SelectedTreeViewNode(tree)) = Null)
							EnableGadget(accept)
						Else
							DisableGadget(accept)
						EndIf
					Case accept
						tf.format = GetFormatByIndex(SelectedTreeViewNode(tree))
						If(tf <> Null) format = tf\format
						Exit
				End Select
			Case $803
				format = 0
				Exit
		End Select
	Wend
	
	For tf.format = Each format
		If(tf\format <> format) snd_format_free(tf\format)
		Delete tf
	Next
	
	Delete Each format_tag
	
	FreeGadget(window)
	Return(format)
End Function