;----------------------------------------------
; snd_voip.bb
;----------------------------------------------
;	This example creates a small program which allows 2 people to connect to
;	eachother and talk in realtime.
;----------------------------------------------
;	Todo:
;		- Add ability to select codecs (or atleast quality)
;		- Proper realtime volume bar
;		- Better error trapping
;		- Optimize
;		- Support multiple devices
;		- Prebuffering
;	Bugs:
;		- Blitz halts execution on Menu dropdowns, mixed with no prebuffering.
;		- Crashes sometimes, most likely a memory leak or device stopped while
;		  working.
;		- Weird noticable breaks between audio sometimes, seems odd because the
;		  recording is on another thread and never buffer underuns. Possibly a block
;		  or compression issue with SndLib.
AppTitle("BlitzVoice")

Include "snd_shared.bb"
Include "snd_guiextras.bb"

;----------------------------------------------
;	Load Speex codec (www.speex.org, THEY ROCK!)
;----------------------------------------------
Global codec = snd_codec_add("speexw.acm")
If(codec = 0)
	Notify("Codec 'speexw.acm' could not be loaded", 1)
	End
EndIf

;------------------------------------
;	Network / Sound
;------------------------------------
Global snd_in, snd_out, snd_encode, snd_decode
Global format_in, format_out, format_net
Global sends_per_second%
Global remote_ip$, remote_port%
Global local_port%
Global tcp_server, tcp_client
Global tcp_server_port = 7868
Global block_size
Global connected
Global msg_id, msg_len

format_out = snd_format_create(1, 8000, 16, 1)
format_in = snd_format_create(1, 8000, 16, 1)
;format_net = FindFormat(41225, 8000, 16, 1, 1000)	;1kb/sec
format_net = FindFormat(41225, 8000, 16, 1, 3075)	;3kb/sec
  
If(format_net = 0)
	Notify("Could not find format!", 1)
	End
EndIf

sends_per_second = 4
block_size = snd_format_bps(format_in) / sends_per_second		; / Times Per Second (Sending rate)
snd_in = snd_in_open(format_in, block_size, 3)
snd_out = snd_out_open(format_out)

;Conversion streams for encoding from formats
snd_encode = snd_convert_open(format_in, format_net)
snd_decode = snd_convert_open(format_net, format_out)

Global data_send = 0
Global data_recv = 0
Global buffer_in = CreateBank(block_size)
Global buffer_fft = CreateBank(1024 * 4)
Global block_size_encode = snd_convert_getsize(snd_encode, block_size)
Global block_size_decode = snd_convert_getsize(snd_decode, block_size_encode)
Global buffer_encode = CreateBank(block_size_encode)
Global buffer_decode = CreateBank(block_size_decode)

tcp_server = CreateTCPServer(7868)
If(tcp_server=0)
	;Notify("Unable to host TCP server! Using random port")
	tcp_server_port = Rand(7869, 7899)
	tcp_server = CreateTCPServer(tcp_server_port)
EndIf

TCPTimeouts 2000, 0

Function ResizeForm()
	Local width = ClientWidth(window)
	Local height = ClientHeight(window)
	
	SetGadgetShape(ChatArea, 4, 12, ClientWidth(window)-8, ClientHeight(window)-144)
	SetGadgetShape(WaveForms[0]\canvas, 4, ClientHeight(window)-52, 96, 24)
	SetGadgetShape(WaveForms[1]\canvas, ClientWidth(window)-100, ClientHeight(window)-52, 96, 24)
	SetGadgetShape(VolumeBars[0]\canvas, 102, ClientHeight(window)-52, 16, 24)
	SetGadgetShape(VolumeBars[1]\canvas, ClientWidth(window)-118, ClientHeight(window)-52, 16, 24)
	SetGadgetShape(ChatField, 4, ClientHeight(window)-24, ClientWidth(window)-8, 20)
	;SetGadgetShape(NetworkGraph\canvas, 4, ClientHeight(window)-128, ClientWidth(window)-8, 64)
End Function

;----------------------------------------------
;	Define Interface
;----------------------------------------------
Global Window = CreateWindow("BlitzVoice [SndLib Demo]", 320, 240, 256, 256, 0, 1 + 8 + 2)
Global ChatArea = CreateTextArea(0, 0, 0, 0, window)
Global WaveForms.WaveForm[2]
	WaveForms[0] = CreateWaveForm(0, 0, 96, 24, window)
	WaveForms[1] = CreateWaveForm(0, 0, 96, 24, window)
Global VolumeBars.VolumeBar[2]
	VolumeBars[0] = CreateVolumeBar(0, 0, 16, 24, window)
	VolumeBars[1] = CreateVolumeBar(0, 0, 16, 24, window)
Global ChatField = CreateTextField(0, 0, 0, 0, window)
;Global NetworkGraph.NetworkGraph = CreateNetworkGraph(0, 0, ClientWidth(window)-8, 64, window)

SetGadgetLayout ChatArea, 1, 0, 1, 0
SetGadgetLayout WaveForms[0]\canvas, 1, 0, 1, 0
SetGadgetLayout WaveForms[1]\canvas, 1, 0, 1, 0
SetGadgetLayout VolumeBars[0]\canvas, 1, 0, 1, 0
SetGadgetLayout VolumeBars[0]\canvas, 1, 0, 1, 0
SetGadgetLayout ChatField, 1, 0, 1, 0
;SetGadgetLayout NetworkGraph\canvas, 1, 0, 1, 0

SetMinWindowSize(Window, 256, 256)
DisableGadget(chatarea)
DisableGadget(chatfield)

SetStatusText(window, "Listening with "+DottedIP(HostIP(CountHostIPs("")))+":"+tcp_server_port)

Global menu_network[8]
	menu_network[0] = CreateMenu("Network", 1, WindowMenu(window))
		menu_network[1] = CreateMenu("Connect", 10, menu_network[0])
		menu_network[2] = CreateMenu("Disconnect", 11,menu_network[0])
		menu_network[3] = CreateMenu("", 0, menu_network[0])
		menu_network[4] = CreateMenu("Exit", 12, menu_network[0])
		
		DisableMenu(menu_network[2])
UpdateWindowMenu(window)

ResizeForm()

;----------------------------------------------
;	Main Loop
;----------------------------------------------
While(True)
	If(WaitEvent(20))
		Select(EventID())
			Case $803
				Disconnect(True)
				Exit
			Case $802
				If(EventSource() = Window)
					ResizeForm()
				EndIf
			Case $401
				Select(EventSource())
					Case chatfield
						If((EventData() = 13) Or (EventData() = 10))
							AddTextAreaText(chatarea, Chr(10)+"<local>"+TextFieldText(chatfield))
							length = Len(TextFieldText(chatfield)) + 4
							WriteByte(tcp_client, 2)
							WriteInt(tcp_client, length)
							WriteString(tcp_client, TextFieldText(chatfield))
							data_send = data_send + Len(TextFieldText(chatfield)) + 5
							
							SetGadgetText(chatfield, "")
						EndIf
				End Select
			Case $1001
				Select(EventData())
					Case 10
						Connect(window)
					Case 11
						Disconnect()
					Case 12
						Exit
				End Select
		End Select
	EndIf
	
	If(connected)
		If(Eof(tcp_client))
			tcp_client = 0
			Disconnect()
		Else
			If(msg_id=0)
				If(ReadAvail(tcp_client) >= 5)
					msg_id = ReadByte(tcp_client)
					msg_len = ReadInt(tcp_client)
					data_recv = data_recv + msg_len + 5
				EndIf
			Else
				If(ReadAvail(tcp_client) >= msg_len)
					Select(msg_id)
						Case 1
							src = CreateBank(msg_len)
							dest = CreateBank(snd_convert_getsize(snd_decode, msg_len))
							ReadBytes(src, tcp_client, 0, msg_len)
							bytes = snd_convert_encode(snd_decode, src, 0, BankSize(src), dest, 0, BankSize(dest))
							snd_out_write(snd_out, dest, 0, bytes)
							
							analyze(dest, buffer_fft)
							DrawWaveForm(WaveForms[1], buffer_fft)
							DrawVolumeBar(VolumeBars[1], buffer_fft)
							
							FreeBank(src)
							FreeBank(dest)
						Case 2
							s$ = ReadString(tcp_client)
							AddTextAreaText(chatarea, Chr(10)+"<remote>"+s)
					End Select
					
					msg_id = 0
					msg_len = 0
				EndIf
			EndIf
			
			If(snd_in_readavail(snd_in) >= block_size)
				snd_in_read(snd_in, buffer_in, block_size)
				
				analyze(buffer_in, buffer_fft)
				DrawWaveForm(WaveForms[0], buffer_fft)
				DrawVolumeBar(VolumeBars[0], buffer_fft)
				;DrawNetworkGraph(NetworkGraph, data_send, data_recv)
				data_send = 0
				data_recv = 0
				
				If(snd_in_readavail(snd_in) > block_size)
					snd_in_flush(snd_in)
				EndIf
				
				encoded_size = snd_convert_encode(snd_encode, buffer_in, 0, block_size, buffer_encode, 0, block_size_encode)
				
				WriteByte(tcp_client, 1)
				WriteInt(tcp_client, encoded_size)
				WriteBytes(buffer_encode, tcp_client, 0, encoded_size)
				data_send = data_send + 5 + encoded_size
			EndIf
		EndIf
	Else
		If(AcceptConnection())
			StartConversation()
		EndIf
	EndIf
Wend

;----------------------------------------------
;	Process the signal, FFT for use with volume
;	and waveform graphs
;----------------------------------------------
Function Analyze(buffer, fft)
	For i=0 To 1023
		index = Float(Float(i) / Float(1023)) * (BankSize(buffer)-2)
		index = (index/2) * 2
		
		value = PeekShort(buffer, index)
		If(value > $7FFF) value = $7FFF-value
		
		PokeFloat(fft, i * 4, value)
	Next
	
	snd_fft(fft, 1024)
End Function

Function AcceptConnection()
	Local stream = AcceptTCPStream(tcp_server)
	If(stream=0) Return(0)
	
	tcp_client = stream
	Return(1)
End Function

;Present a dialog, attempt to connect to a remote host.
Function Connect(parent%=0)
	Local x = GadgetX(parent) + GadgetWidth(parent)/2 - 176/2
	Local y = GadgetY(parent) + GadgetHeight(parent)/2 - 142/2
	Local window = CreateWindow("Connect", x, y, 176, 142, parent, 1)
	Local static[4]
		static[0] = CreateLabel("Remote IP", 8, 8, 64, 20, window)
		static[1] = CreateLabel("Remote Port", 8, 32, 64, 20, window)
		static[2] = CreateLabel("Local Port", 8, 54, 64, 20, window)
		static[3] = CreateLabel("", 4, 100, 162, 20, window)
	Local textfield[3]
		textfield[0] = CreateTextField(80, 8, 80, 20, window)
		textfield[1] = CreateTextField(80, 30, 80, 20, window)
		textfield[2] = CreateTextField(80, 52, 80, 20, window)
		
		SetGadgetText(textfield[0], "127.0.0.1")
		SetGadgetText(textfield[1], "7868")
		SetGadgetText(textfield[2], "0")
		
	Local button = CreateButton("Connect", 4, 76, 162, 20, window)
	Local accepted = False
	Local connecting = False
	
	If(parent) DisableGadget(parent)
	
	While(True)
		If(WaitEvent(60))
			Select(EventID())
				Case $803
					accepted = False
					Exit
				Case $401
					Select(EventSource())
						Case button
							If(connecting)
								SetGadgetText(button, "Connect")
								SetGadgetText(static[3], "")
								connecting = False
							Else
								SetGadgetText(button, "Abort")
								SetGadgetText(static[3], "Connecting...")
								connecting = True
								
								tcp_client = OpenTCPStream(TextFieldText(textfield[0]), TextFieldText(textfield[1]), TextFieldText(textfield[2]))
								If(tcp_client=0)
									SetGadgetText(button, "Connect")
									SetGadgetText(static[3], "Failed to connect")
									accepted = False
								Else
									StartConversation()
									accepted = True
									Exit
								EndIf
							EndIf
					End Select
			End Select
		EndIf
	Wend
	
	If(parent) EnableGadget(parent)
	
	FreeGadget(window)
	Return(accepted)
End Function

Function DrawVolumeBar(vb.VolumeBar, bank)
	width = ClientWidth(vb\canvas)
	height = ClientHeight(vb\canvas)
	SetBuffer(CanvasBuffer(vb\canvas))
	
	ColorGui(COLOR_BTNFACE)
	Rect 0, 0, width, height
	
	If(bank <> 0)
		total# = 0
		vb\peak = vb\peak - .1
		
		For i=0 To 512
			total = total + PeekFloat(bank, i*4)
		Next
		
		p# = ((total / 512) / $7FFF) / height
		If(p > vb\peak) vb\peak = p
		
		Color 0, 255, 0
		Rect 0, height - (height * p), width, height * p
		
		Color 255, 0, 0
		Rect 0, height - (height * vb\peak), width, 1
	EndIf
	
	ColorGui(COLOR_WINDOWFRAME)
	Rect 0, 0, width, 1
	Rect 0, 0, 1, height
	
	ColorGui(COLOR_BTNHIGHLIGHT)
	Rect 0, height-1, width, 1
	Rect width-1, 0, 1, height
	
	FlipCanvas(vb\canvas)
End Function

Function DrawWaveForm(wf.WaveForm, bank)
	width = ClientWidth(wf\canvas)
	height = ClientHeight(wf\canvas)
	SetBuffer(CanvasBuffer(wf\canvas))
	
	ColorGui(COLOR_BTNFACE)
	Rect 0, 0, width, height
	
	If(bank <> 0)		
		For i=0 To width-1
			p# = Float(i) / Float(width-1)
			index = (p * 512)		
			f# = PeekFloat(bank, index * 4)
			f = (f / $FFFF) / height
			
			Color 0, 255,0		
			Rect width*p, height - (height*f), 1, (height*f)
			
			wf\peaks[index] = wf\peaks[index] - .1
			If(f > wf\peaks[index]) wf\peaks[index] = f
			
			WritePixel width*p, height-(height*wf\peaks[index]), $FF0000
		Next
	EndIf
	
	ColorGui(COLOR_WINDOWFRAME)
	Rect 0, 0, width, 1
	Rect 0, 0, 1, height
		
	ColorGui(COLOR_BTNHIGHLIGHT)
	Rect 0, height-1, width, 1
	Rect width-1, 0, 1, height
	
	FlipCanvas(wf\canvas)
End Function

Function StartConversation()
	EnableGadget(chatarea)
	EnableGadget(chatfield)
	DisableMenu(menu_network[1])
	EnableMenu(menu_network[2])
	connected = True
	
	UpdateWindowMenu(window)
	
	snd_in_start(snd_in)
End Function

Function Disconnect(server%=False)
	If(tcp_client)
		CloseTCPStream(tcp_client)
		tcp_client = 0
	EndIf
	
	If((tcp_server <> 0) And (server <> 0))
		CloseTCPServer(tcp_server)
		tcp_server = 0
	EndIf
	
	DisableGadget(chatarea)
	DisableGadget(chatfield)
	EnableMenu(menu_network[1])
	DisableMenu(menu_network[2])
	connected = False
	
	UpdateWindowMenu(window)
	
	DrawWaveForm(WaveForms[0], 0)
	DrawWaveForm(WaveForms[1], 0)
	
	snd_in_stop(snd_in)
End Function

;----------------------------------------------
; Custom Control Vodoo
;----------------------------------------------

Type NetworkGraph
	Field Highest
	Field Canvas
End Type

Type VolumeBar
	Field Canvas
	Field Peak#
End Type

Type WaveForm
	Field Canvas
	Field Peaks#[512]
End Type

Function CreateNetworkGraph.NetworkGraph(x%, y%, width%, height%, window%)
	Local ng.NetworkGraph = New NetworkGraph
	
	ng\canvas = CreateCanvas(x, y, width, height, window)
	width = ClientWidth(ng\canvas)
	height = ClientHeight(ng\canvas)
	SetBuffer(CanvasBuffer(ng\canvas))
	
	ColorGui(COLOR_BTNFACE)
	Rect 0, 0, width, height
		
	ColorGui(COLOR_WINDOWFRAME)
	Rect 0, 0, width, 1
	Rect 0, 0, 1, height
	
	ColorGui(COLOR_BTNHIGHLIGHT)
	Rect 0, height-1, width, 1
	Rect width-1, 0, 1, height
	
	Return(ng)
End Function

Function DrawNetworkGraph(ng.NetworkGraph, send%, recv%)
	width = ClientWidth(ng\canvas)
	height = ClientHeight(ng\canvas)
	SetBuffer(CanvasBuffer(ng\canvas))
		
	If(send > ng\highest) ng\highest = send
	If(recv > ng\highest) ng\highest = recv
	
	CopyRect 1, 1, width-1, height-1, 0, 1
	
	ColorGui(COLOR_BTNFACE)
	Rect width-1, 0, 1, height
	
	Color 255, 255, 0
	p# = (Float(recv) / Float(ng\highest))	
	Rect width-1, height-(p*height), 1, p*height
	
	Color 0, 255, 0
	p# = (Float(send) / Float(ng\highest))	
	Rect width-1, height-(p*height), 1, p*height
		
	ColorGui(COLOR_WINDOWFRAME)
	Rect 0, 0, width, 1
	Rect 0, 0, 1, height
	
	ColorGui(COLOR_BTNHIGHLIGHT)
	Rect 0, height-1, width, 1
	Rect width, 0, 1, height
		
	FlipCanvas(ng\canvas)
End Function

Function CreateVolumeBar.VolumeBar(x%, y%, width%, height%, window%)
	Local vb.VolumeBar = New VolumeBar
	
	vb\canvas = CreateCanvas(x, y, width, height, window)
	vb\peak = 0.0
	DrawVolumeBar(vb, 0)
	
	Return(vb)
End Function

Function CreateWaveForm.WaveForm(x%, y%, width%, height%, window)
	Local wf.WaveForm= New WaveForm
	
	wf\canvas = CreateCanvas(x, y, width, height, window)
	DrawWaveForm(wf, 0)
	
	Return(wf)
End Function